;;; x-face-mule.el -- show X-Face in message buffer for text based Emacs/MULE

;;; Original is also named x-face-mule.el, part of bitmap-mule package.

;;; Copyright (C) 1997 KORIYAMA Naohiro
;;; Author: KORIYAMA Naohiro <kory@ba2.so-net.or.jp>
;;; Version: 0.12
;;; Created: 1997/10/24
;;; Revised: 1997/12/12
;;; Keywords: X-Face, bitmap, Emacs, MULE

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2, or (at your option)
;;; any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with GNU Emacs; see the file COPYING.  If not, write to the
;;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;;; Boston, MA 02111-1307, USA.

;;; [USAGE]
;;; 1. "bitmap-mule" package and related packages or
;;;    "bitmap.tar.gz" (etl version bitmap.el) are required.
;;;    also "compface.tar.gz" (uncompface) is required.
;;;
;;; 2. Setting
;;;    [before Mew 1.90]
;;;    add the following in your .emacs,
;;;
;;;    (if window-system
;;;       (progn
;;;          (add-hook 'mew-summary-display-message-filter-hook ;; 1.90
;;;     		'x-face-mule:x-face-decode-message-header)
;;;          (add-hook 'mew-message-hook ;; 1.70 or older
;;;     		'x-face-mule:x-face-decode-message-header)))
;;;    (require 'x-face-mule)
;;;
;;;    [Mew 1.92 or later]
;;;    add the following in your .emacs, 
;;;
;;;    (if window-system
;;;        (progn
;;;          (setq mew-opt-highlight-x-face t)
;;;          (setq mew-opt-highlight-x-face-function
;;;     	    (function
;;;     	     (lambda (beg end)
;;;	              (x-face-mule:x-face-decode-message-header beg end))))))
;;;    (require 'x-face-mule)
;;;
;;;    WARNING: MAYBE YOU WILL HAVE TO CHANGE from mew-opt-*  to mew-use-* 
;;;             or Mew-* , SO CHECK which of these variable is used in Mew.
;;;
;;;    [tm-MUA] i.e. Gnus5(or later),VM,mh-e,cmail
;;;    add the following in your .emacs
;;;    (add-hook 'mime-viewer/content-header-filter-hook
;;;              'x-face-mule:gnus-art-mime-highlight-headers t)
;;;    (require 'x-face-mule)
;;;
;;;    [SEMI-MUA] i.e. Gnus5(or later),VM,mh-e,cmail
;;;    add the following in your .emacs
;;;    (add-hook 'mime-view-content-header-filter-hook
;;;              'x-face-mule:gnus-art-mime-highlight-headers t)
;;;    (require 'x-face-mule)
;;;
;;; 3. that's all!
;;;
;;; [CUSTOMIZATION]
;;; 1. If you don't want to show X-Face at "From:" field,
;;;    add the following in your .emacs
;;;    (setq x-face-mule:highlight-x-face-position 'x-face)
;;;    and you can show X-Face at "X-Face:" field.
;;;
;;; 2. If you want to show X-Face like XEmacs style ,
;;;    add following code in your .eamcs.
;;;    (setq x-face-mule:highlight-x-face-style 'xmas)
;;;
;;;    But, this feature won't work well on some MUA...
;;;
;;; [USER FUNCTION]
;;; (x-face-mule:toggle-x-face-position) toggle show method.
;;;                                    from->x-face->nil->from...
;;;
;;; [TODO]
;;; 1. cool code!
;;; 2. and many many things
;;;
;;; [THANKS TO]
;;; OKUNISHI Fujikazu <fuji0924@mbox.kyoto-inet.or.jp>
;;; Yuuichi Teranishi <teranisi@isl.ntt.co.jp>
;;; TSUMURA Tomoaki <tsumura@kuis.kyoto-u.ac.jp>
;;; Shiono@FSC <jun@fsc.fsc.fs.fujitsu.co.jp>
;;; Kazuhiro Ohta <ohta@ele.cst.nihon-u.ac.jp>
;;; MORIOKA Tomohiko <morioka@jaist.ac.jp>
;;; Tatsuya Ichikawa <ichikawa@hv.epson.co.jp>
;;; Katsumi Yamaoka <yamaoka@ga.sony.co.jp>
;;;

;;; Code:
(require 'bitmap)

(or (and (condition-case ()
	     (require 'custom)
	   (file-error nil))
	 (fboundp 'defgroup)
	 (fboundp 'defcustom))
    (progn
      (require 'backquote)
      (defmacro defgroup (&rest args))
      (defmacro defcustom (symbol value &optional doc &rest args)
	(` (defvar (, symbol) (, value) (, doc))))
      ))


(defconst x-face-mule:version "0.12")

;;; Silence the byte compiler.
(eval-when-compile
  (mapcar
   (function (lambda (sym) (eval (list 'defvar sym nil))))
   '(
     mime-view-original-major-mode
     mime::preview/original-major-mode
     ))
  )


(defgroup x-face-mule nil
  "show X-Face inline for Emacs/Mule."
  :prefix "x-face-mule:"
  :group 'news
  :group 'mail)

(defcustom x-face-mule:uncompface-program "uncompface"
  "*Program for decoding X-Face stirg to UN|X Icon."
  :group 'x-face-mule
  :type 'file)

(defcustom x-face-mule:highlight-x-face-position 'from
  "*Where to show X-Face.
`from' at From: field, `x-face' at X-Face: field, `off' don't show."
  :group 'x-face-mule
  :type '(radio (variable-item from)
		(variable-item x-face)
		(variable-item off)))

(defcustom x-face-mule:highlight-x-face-style 'default
  "*Show style of X-Face.
when `default' is set, show like this:
From:xxxxxx
     xxxxxx
     xxxxxx Foo Bar <foobar@someware.org>

when `xmas' is set, show like this:
     xxxxxx
     xxxxxx
From:xxxxxx Foo Bar <foobar@someware.org>"
  :group 'x-face-mule
  :type '(radio (variable-item default)
		(variable-item xmas)))

(defcustom x-face-mule:highlight-x-face-position-alist nil
  "* If non nil, x-face-position is changed by major-mode.
A format is like this:

(setq x-face-mule:highlight-x-face-position-alist
      '((mew-message-mode . from) 
	(mime-view-mode . x-face)))"
  :group 'x-face-mule
  :type '(repeat (cons (symbol :tag "MAJOR-MODE")
		       (radio :tag "POSITION"
			      (variable-item from)
			      (variable-item x-face)
			      (variable-item off)))))

(defcustom x-face-mule:highlight-x-face-style-alist nil
  "* If non nil, x-face-style is changed by major-mode.
A format is like this:

(setq x-face-mule:highlight-x-face-style-alist
      '((mew-message-mode . xmas)
	(mime-view-mode . default)))"
  :group 'x-face-mule
  :type '(repeat (cons (symbol :tag "MAJOR-MODE")
		       (radio :tag "STYLE"
			      (variable-item default)
			      (variable-item xmas)))))

(defcustom x-face-mule:highlight-x-face-refresh-method-alist
  '((mew-summary-mode  . (mew-summary-display nil))
    (mew-virtual-mode  . (mew-summary-display nil))
    (gnus-summary-mode . (gnus-summary-select-article nil t))
    )
  "* Alist for refreshing message-buffer.
Format:
\    '((MAJOR-MODE . (FUNCTION ARGS)))"
  :group 'x-face-mule
  :type '(repeat (cons (symbol :tag "MAJOR-MODE")
		       (function :tag "FUNCTION"))))

(defcustom x-face-mule:x-face-decode-message-header-hook nil
  "*Hook called before decoding X-Face field."
  :group 'x-face-mule
  :type 'hook)

;;; highlight face for X-Face
(if (fboundp 'defface)
    (defface x-face-mule:highlight-x-face
      '((((class color)
	  (background dark))
	 (:foreground "Black"  :background "White"))
	(((class color)
	  (background light))
	 (:foreground "Black"))
	(t
	 ()))
      "X-Face face."
      :group 'x-face-mule)
  (or (internal-find-face 'x-face-mule:highlight-x-face)
      (progn
	(make-face 'x-face-mule:highlight-x-face)
	(set-face-foreground 'x-face-mule:highlight-x-face "Black")
	(set-face-background 'x-face-mule:highlight-x-face "White")
	)))

(setq x-face-mule:highlight-x-face
      (if (boundp 'x-face-mule:highlight-x-face)
	  (symbol-value 'x-face-mule:highlight-x-face)
	'x-face-mule:highlight-x-face))

;;;
;;; emulate `bitmap-mule/x-face-mule.el'
;;;
(defvar uncompface-program x-face-mule:uncompface-program)
(defalias 'x-face-decode-message-header
  'x-face-mule:x-face-decode-message-header)

;;; toggle x-face-position
(defun x-face-mule:toggle-x-face-position (&optional arg)
  (interactive "P")
  (let ((pos (cond ((eq x-face-mule:highlight-x-face-position 'from)
		    'x-face)
		   ((eq x-face-mule:highlight-x-face-position 'x-face)
		    'off)
		   (t 'from)))
	(table '(("From:" . from) ("X-Face:" . x-face) ("Off" . off)))
	(method (cdr
		 (assq major-mode
		       x-face-mule:highlight-x-face-refresh-method-alist)))
	msg)
    (setq x-face-mule:highlight-x-face-position
	  (if arg
	      (cdr (assoc (completing-read
			   "Show X-Face at> " table nil t
			   (concat (capitalize (symbol-name pos))
				   (if (memq pos '(from x-face)) ":")))
			  table))
	    pos))
    (and (interactive-p)
	 (setq msg (cond
		    ((eq x-face-mule:highlight-x-face-position 'from)
		     "Show X-Face at From:")
		    ((eq x-face-mule:highlight-x-face-position 'x-face)
		     "Show X-Face at X-Face:")
		    (t "Don't Show X-Face")))
	 (message "%s..." msg))
    (and (fboundp (car method)) (apply 'funcall method))
    (and (interactive-p)
	 (message "%s...done" msg))))

;;; x-face conversion cache
(defvar x-face-mule:x-face-to-rectangle-cache nil)

(defun x-face-mule:x-face-convert-x-face-to-rectangle (string)
  "convert x-face string to rectangle using cache."
  (let ((data (assoc string x-face-mule:x-face-to-rectangle-cache)))
    (if (not data)
	(progn
	  (setq data (cons
		      string
		      (x-face-mule:x-face-convert-icon-to-rectangle
		       (x-face-mule:x-face-convert-x-face-to-icon string)
		       )))
	  (setq x-face-mule:x-face-to-rectangle-cache
		(cons data x-face-mule:x-face-to-rectangle-cache))))
    (cdr data)))

;;;
;;; originate from 'x-face-decode-message-header () in x-face-mule.el
;;;
(defun x-face-mule:x-face-convert-x-face-to-icon (string)
  "decode x-face string to UN|X ICON."
  (save-excursion
    (let ((tmp-buffer (get-buffer-create "*x-face-tmp*"))
	  ret-string)
      (set-buffer tmp-buffer)
      (insert string)
      (call-process-region (point-min) (point-max)
			   x-face-mule:uncompface-program t t nil)
      (setq ret-string (buffer-substring (point-min) (point-max)))
      (kill-buffer tmp-buffer)
      ret-string
      )))

(defun x-face-mule:x-face-convert-vector-to-rectangle (vector)
  "make x-face rectangle from vector."
  (let ((ret nil)
	line i k k+6)
    (setq k 0)
    (setq i 0)
    (while (< i 3)
      (setq line "")
      (setq k (* i 6)  k+6 (+ k 6))
      (while (< k k+6)
	(setq line (concat line (bitmap-compose (aref vector k))))
	(setq k (1+ k))
	)
      (setq ret (append ret (list line)))
      (setq i (1+ i)))
    ret
    ))

(defun x-face-mule:x-face-convert-icon-to-rectangle (icon)
  "decode UN|X ICON to rectangle."
  (save-excursion
    (let ((tmp-buffer (get-buffer-create "*x-face-tmp*"))
	  i temp cmp k k+6)
      (set-buffer tmp-buffer)
      (insert icon)
      (goto-char (point-min))
      (search-forward "0x" nil t)
      (setq cmp (make-vector 18 nil))
      (setq i 0)
      (while (< i 48)
	(setq k (* (/ i 16) 6))
	(setq k+6 (+ k 6))
	(while (< k k+6)
	  (setq temp (buffer-substring (point)
				       (+ (point) 2)))
	  (aset cmp k (concat (aref cmp k) temp))
	  (setq k (1+ k))
	  (setq temp (buffer-substring (+ (point) 2)
				       (+ (point) 4)))
	  (aset cmp k (concat (aref cmp k) temp))
	  (setq k (1+ k))
	  (search-forward "0x" nil t))
	(setq i (1+ i))
	)
      (kill-buffer tmp-buffer)
      (x-face-mule:x-face-convert-vector-to-rectangle cmp))))

(defun x-face-mule:x-face-insert-at-point (rectangle)
  "insert x-face rectangle and overlay its face."
  (let ((insertcolumn (current-column))
	(first t)
	(gnusp (memq (or (and (boundp 'mime-view-original-major-mode) ; SEMI
			      mime-view-original-major-mode)
			 (and (boundp 'mime::preview/original-major-mode) ; tm
			      mime::preview/original-major-mode)
			 major-mode)
		     '(gnus-original-article-mode gnus-article-mode)))
	beg-point)
    (while rectangle
      (or first
	  (progn
	    (forward-line 1)
	    (or (bolp) (insert ?\n))
	    (if (fboundp 'move-to-column-strictly)
		(move-to-column-strictly insertcolumn t) ; XEmacs only ??
	      (move-to-column insertcolumn t)) ; text Emacs
	    ))
      (setq first nil)
      (setq beg-point (point))
      (insert (car rectangle))
      (if gnusp
	  (put-text-property beg-point (point) 'x-face-mule:bitmap-image t)
	(overlay-put (make-overlay beg-point (point))
		     'face x-face-mule:highlight-x-face))
      (setq rectangle (cdr rectangle)))))

(defun x-face-mule:insert-invisible-from ()
  (let (beg)
    (setq beg (point))
    (insert "From:")
    (put-text-property beg (point) 'invisible t)
    (insert "     ")))

(defun x-face-mule:x-face-allocate-lines (beg end height)
  "allocate new lines according to x-face-mule:highlight-x-face-position.
returns the begin-point of the x-face rectangle."
  (let (begin-point
	(n 1))
    (cond
     ((eq x-face-mule:highlight-x-face-position 'from)
      (goto-char beg)
      (if (re-search-forward  "^From:" end t)
	  (progn
	    (cond
	     ((eq x-face-mule:highlight-x-face-style 'xmas)
	      (beginning-of-line)
	      (x-face-mule:insert-invisible-from)
	      (setq begin-point (point))
	      (insert "\n")
	      (x-face-mule:insert-invisible-from)
	      (insert "\n")
	      (while (< n height)
		(x-face-mule:insert-invisible-from)
		(insert "\n")
		(x-face-mule:insert-invisible-from)
		(insert "\n")
		(x-face-mule:insert-invisible-from)
		(insert "\n")
		(setq n (1+ n))))
	     (t ; 'default
	      (setq begin-point (point))
	      (insert "\n     \n     ")
	      (while (< n height)
		(insert "\n     \n     \n     ")
		(setq n (1+ n))))))))
     ((eq x-face-mule:highlight-x-face-position 'x-face)
      (insert "X-Face: ")
      (setq begin-point (point))
      (insert "\n        \n        ")
      (while (< n height)
	(insert "\n        \n        \n        ")
	(setq n (1+ n)))
      (insert "\n")))
    begin-point ;; return value.
    ))

(defun x-face-mule:change-highlight-x-face-method-by-alist ()
  (if x-face-mule:highlight-x-face-position-alist
      (let ((position (cdr
		       (assq major-mode
			     x-face-mule:highlight-x-face-position-alist))))
	    (if (memq position '(from x-face off))
		(setq x-face-mule:highlight-x-face-position position))
	))
  (if x-face-mule:highlight-x-face-style-alist
      (let ((style (cdr
		    (assq major-mode
			  x-face-mule:highlight-x-face-style-alist))))
	    (if (memq style '(default xmas))
		(setq x-face-mule:highlight-x-face-style style))
	)))

(defun x-face-mule:analyze-x-face-geometry (beg end)
  (goto-char beg)
  (if (re-search-forward
       "^X-Face-Type: geometry=[0-9]+x[0-9]+" end t)
      (let ((sp (+ 22 (match-beginning 0))) (ep (match-end 0)))
	(buffer-substring sp ep))
    (let ((i 0))
      (goto-char beg)
      (while (re-search-forward
	      "^X-Face: *\\(.*\\(\n[ \t].*\\)*\\)\n" end t)
	(setq i (1+ i)))
      (concat i "x1"))))

(defun x-face-mule:x-face-decode-message-header (&optional beg end)
  (run-hooks 'x-face-mule:x-face-decode-message-header-hook)
  (cond ((memq x-face-mule:highlight-x-face-position '(from x-face))
	 (if (null end)
	     (progn
	       (setq beg (point-min))
	       (setq end (if (search-forward "\n\n" nil t)
			     (point) (point-max))) ))
	 (save-restriction
	   (narrow-to-region beg end)
	   (let ((inhibit-read-only t)
		 buffer-read-only
		 x-face faces faces-s first begin-point
		 geometry xp WIDTH HEIGHT w h)
	     (setq geometry (x-face-mule:analyze-x-face-geometry beg end))
	     (setq xp (string-match "x" geometry))
	     (setq WIDTH (string-to-number (substring geometry 0 xp)))
	     (setq HEIGHT (string-to-number (substring geometry (1+ xp))))
	     (goto-char beg)
	     (setq h 0)
	     (while (< h HEIGHT)
	       (setq w 0)
	       (while (< w WIDTH)
		 (re-search-forward
		  "^X-Face: *\\(.*\\(\n[ \t].*\\)*\\)\n" end t)
		 (setq faces-s
		       (cons
			(buffer-substring (match-beginning 1) (match-end 1))
			faces-s
			))
		 (delete-region (match-beginning 0) (match-end 0))
		 (setq w (1+ w)))
	       (setq faces (append faces faces-s))
	       (setq faces-s nil)
	       (setq h (1+ h)))
	     ;;
	     (x-face-mule:change-highlight-x-face-method-by-alist)
	     (setq first t)
	     (setq h 0)
	     (while (< h HEIGHT)
	       (setq w 0)
	       (while (< w WIDTH)
		 (setq x-face
		       (x-face-mule:x-face-convert-x-face-to-rectangle
			(car faces)))
		 (if first ; allocate lines for face field
		     (setq begin-point
			   (x-face-mule:x-face-allocate-lines beg end HEIGHT)))
		 (goto-char begin-point)
		 (x-face-mule:x-face-insert-at-point x-face)
		 (setq faces (cdr faces))
		 (setq first nil)
		 (setq w (1+ w)))
	       (setq h (1+ h))
	       (if (< h HEIGHT)
		   (progn
		     (end-of-line 2)
		     (setq begin-point (point))))
	       )))
	 )))

;;;
;;; MUA dependencies
;;;
;;; for Gnus5
(defcustom x-face-mule:gnus-highlight-headers-anyway nil
  "*If non-nil, display X-Face even if the MIME processing is off."
  :group 'x-face-mule
  :type 'boolean)

(defun x-face-mule:narrow-to-head (&optional widen)
  "Narrow the buffer to the head of the message."
  (and widen (widen))
  (narrow-to-region
   (goto-char (point-min))
   (if (search-forward "\n\n" nil 1)
       (1- (point))
     (point-max)))
  (goto-char (point-min)))

(defun x-face-mule:gnus-article-highlight-x-face ()
  (let ((inhibit-read-only t)
	buffer-read-only st nd column)
    (or (eq 'x-face x-face-mule:highlight-x-face-position)
	(and (eq 'xmas x-face-mule:highlight-x-face-style)
	     (progn
	       (goto-char (point-max))
	       (while (re-search-backward "^From:" nil t)
		 (setq st (match-beginning 0)
		       nd (or (and (re-search-backward "^[^F]" nil t)
				   (match-beginning 0))
			      (point-min)))
		 (goto-char st)
		 (while (re-search-backward "^From:" nd t)
		   (and (get-text-property (+ 5 (match-end 0))
					   'x-face-mule:bitmap-image)
			(overlay-put
			 (make-overlay (match-beginning 0) (match-end 0))
			 'invisible t)))
		 (goto-char nd)))))
    (goto-char (point-min))
    (and (re-search-forward "^From:\\(     \\)?\\|^X-Face: " nil t)
	 (progn
	   (setq st (match-beginning 1))
	   (or
	    (and st
		 (get-text-property st 'x-face-mule:bitmap-image)
		 (goto-char st))
	    (get-text-property (setq st (point)) 'x-face-mule:bitmap-image)))
	 (or (eq x-face-mule:highlight-x-face (get-text-property st 'face))
	     (progn
	       (setq column (current-column)
		     nd (or (and (re-search-forward "^[^\n\t F]\\|^$" nil t)
				 (match-beginning 0))
			    (point-max)))
	       (goto-char st)
	       (while (< (point) nd)
		 (move-to-column column)
		 (and (get-text-property (point) 'x-face-mule:bitmap-image)
		      (put-text-property
		       (point)
		       (next-single-property-change
			(point) 'x-face-mule:bitmap-image)
		       'face x-face-mule:highlight-x-face))
		 (forward-line 1)))))))

(defun x-face-mule:gnus-article-highlight-headers ()
  "Highlight article headers as specified by `gnus-header-face-alist'."
  (let ((ofun (symbol-function 'message-narrow-to-head)))
    (unwind-protect
	(progn
	  (fset 'message-narrow-to-head
		(symbol-function 'x-face-mule:narrow-to-head))
	  (orig::gnus-article-highlight-headers))
      (fset 'message-narrow-to-head ofun))
    (x-face-mule:gnus-article-highlight-x-face)))

(defun x-face-mule:gnus-article-hide-headers (&optional
					      interactive &rest args)
  "Toggle whether to hide unwanted headers and possibly sort them as well.
If given a negative prefix, always show; if given a positive prefix,
always hide."
  (interactive '(t))
  (save-excursion
    (set-buffer gnus-article-buffer)
    (if interactive
	(call-interactively 'article-hide-headers)
      (apply 'article-hide-headers args))
    (x-face-mule:gnus-article-highlight-x-face)))

(defun x-face-mule:gnus-article-hide-boring-headers (&optional
						     interactive &rest args)
  "Toggle hiding of headers that aren't very interesting.
If given a negative prefix, always show; if given a positive prefix,
always hide."
  (interactive '(t))
  (save-excursion
    (set-buffer gnus-article-buffer)
    (let ((inhibit-read-only t)
	  buffer-read-only)
      (goto-char (point-min))
      (while (re-search-forward "^From: *_?$" nil t)
	(insert "_"))
      (if interactive
	  (call-interactively 'article-hide-boring-headers)
	(apply 'article-hide-boring-headers args))
      (goto-char (point-min))
      (while (re-search-forward "^From: *_$" nil t)
	(backward-delete-char 1)))))

(defun x-face-mule:substitute-gnus-highlight-function ()
  (remove-hook 'gnus-art-load-hook
	       'x-face-mule:substitute-gnus-highlight-function)
  (remove-hook 'gnus-article-prepare-hook
	       'x-face-mule:substitute-gnus-highlight-function)
  (or (fboundp 'article-hide-headers)
      (fset 'article-hide-headers
	    (symbol-function 'gnus-article-hide-headers)))
  (or (fboundp 'article-hide-boring-headers)
      (fset 'article-hide-boring-headers
	    (symbol-function 'gnus-article-hide-boring-headers)))
  (or (fboundp 'orig::gnus-article-highlight-headers)
      (fset 'orig::gnus-article-highlight-headers
	    (symbol-function 'gnus-article-highlight-headers)))
  (mapcar
   (function
    (lambda (fun)
      (fset fun (symbol-function (intern (format "x-face-mule:%s" fun))))))
   '(gnus-article-highlight-headers
     gnus-article-hide-headers
     gnus-article-hide-boring-headers))
  (add-hook 'gnus-article-display-hook
	    'x-face-mule:gnus-highlight-headers-if-no-mime t)
  (setq gnus-strict-mime nil)
  (if (stringp gnus-visible-headers)
      (or (string-match gnus-visible-headers "X-Face:")
	  (setq gnus-visible-headers
		(format "\\(%s\\)\\|^X-Face:" gnus-visible-headers)))
    (setq gnus-visible-headers
	  (append gnus-visible-headers '("^X-Face:"))))
  )

(if (featurep 'gnus-art)
    (x-face-mule:substitute-gnus-highlight-function)
  (progn
    (add-hook 'gnus-art-load-hook
	      'x-face-mule:substitute-gnus-highlight-function)
    (add-hook 'gnus-article-prepare-hook
	      'x-face-mule:substitute-gnus-highlight-function)))

(defun x-face-mule:gnus-highlight-headers-if-no-mime ()
  (and window-system
       (eq 'gnus-article-mode major-mode)
       x-face-mule:gnus-highlight-headers-anyway
       (memq x-face-mule:highlight-x-face-position '(from x-face))
       (progn
	 (x-face-mule:x-face-decode-message-header)
	 (x-face-mule:gnus-article-highlight-x-face)
	 )))

(defun x-face-mule:gnus-art-mime-highlight-headers ()
  (and window-system
       (eq 'gnus-original-article-mode
	   (or (and (boundp 'mime-view-original-major-mode)
		    mime-view-original-major-mode)
	       (and (boundp 'mime::preview/original-major-mode)
		    mime::preview/original-major-mode)))
       (>= (point-min) 1)
       (progn
	 (gnus-article-date-local)
	 (article-hide-headers)
	 (gnus-article-highlight-headers)
	 )))

(provide 'x-face-mule)
(provide 'mew-xface-mule) ; :-p

;;; x-face-mule.el ends here
